unit main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms,
  ChipThr, ExtCtrls, Controls ;

type
  TMainForm = class(TForm)
    Book: TNotebook;
    CaretTimer: TTimer;
    Tabs: TPanel;
    TabPad: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormDestroy(Sender: TObject);
    procedure CaretTimerTimer(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure TabsResize(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormActivate(Sender: TObject);
  private
    { Private declarations }
    Threads:TList;
    TabNames:TStrings;
    procedure BasTerminated(Sender:TObject);
    procedure InputKey(Index:integer; var Key: Char);
    procedure OutputKey(Sender:TObject; Key:Char);
    procedure BasIdle(Sender:TObject);
    function GetActiveMachine:integer;
    procedure SetActiveMachine(value:integer);
    procedure ConsoleStateChanged(Sender:TObject; repaint:boolean);
    procedure RestartBlinker;
    procedure ActivateThread(index:integer);
    procedure AddTabDraw(incr:integer);
    procedure ScrollActiveConsole(dif:integer);
    procedure CheckScrollLock(query:boolean);
  public
    procedure PauseActiveMachine;
    function NewMachine(const cmd:string; active:boolean; hint:integer):integer;
    function ActiveThread:TObject;
    function ActiveView:TObject;
    property ActiveMachine:integer read GetActiveMachine write SetActiveMachine;
  end;

var
  MainForm: TMainForm;

implementation
uses basShell,Chip,ConsoleApp;

{$R *.DFM}

type
 TBas=class(TBasicThread)
 public
  function Shell(const cmd:string):integer;
 end;

 TBasShell=class(TExecutor)
 protected
  thr:TBasicThread;
 protected
  function ReadChar(Process: THandle; var Buffer:Char):boolean;override;
 public
  constructor Create(AThread:TBasicThread);
 end;

 constructor TBasShell.Create;
 begin
  inherited Create;
  Script:=AThread.Script;
  thr:=AThread
 end;

 function TBasShell.ReadChar;
 begin
  // TODO: !!!!
  Result:=thr.KeyPressed;
  if Result then Buffer:=thr.ReadKey
 end;

 function TBas.Shell;
 var exec:TExecutor;
 begin
  exec:=TBasShell.Create(Self);
  try
   Result:=exec.Execute(cmd)
  finally
   exec.Free
  end
 end;

{$DEFINE SCREENBUF}

const ROWS=25;
{$IFDEF SCREENBUF}
const BUFFERROWS=ROWS+(39*ROWS);    // TUNE:
const MAXBUFFERROW=BUFFERROWS-1;
const BASEBUFROW=BUFFERROWS-ROWS;
{$ELSE}
const BUFFERROWS=ROWS;
const MAXBUFFERROW=BUFFERROWS-1;
{$ENDIF}
const COLS=80;
const MAXROW=ROWS-1;
const MAXCOL=COLS-1;
const ROWFLAGS=1;

var FCaretState:boolean=false;
var FScrollLock:boolean=false;

type
 TConsoleStateChangedEvent=procedure (Sender:TObject; repaint:boolean) of object;
 TScreen=array [0..MAXBUFFERROW,0..MAXCOL+ROWFLAGS] of char;
 PScreen=^TScreen;
 TConsole=class(TPanel)
 protected
  FOnStateChanged:TConsoleStateChangedEvent;
  FModified:boolean;
  FActive:boolean;
  FBusy:boolean;
  FCaretVisible:boolean;
  FCharSize:TPoint;
  FOrigin:TPoint;
  {$IFDEF SCREENBUF}
  minBufRow,bufRow:integer;
  {$ENDIF}
  FScreen:PScreen;
  row,col:integer;
  procedure Paint;override;
  function Line(r:integer):string;
  procedure HideCaret;
  procedure ShowCaret;
  procedure DrawCaret(invert:boolean);
  procedure SetCaretVisible(value:boolean);
  procedure StateChanged(repaint:boolean);
  procedure PaintScrollers;
  procedure OpenScreen;
 public
  procedure SetActive(value:boolean; immediately:boolean);
  procedure SetBusy(value:boolean; immediately:boolean);
  procedure SetModified(value:boolean; immediately:boolean);
 public
  destructor Destroy;override;
  procedure FontChanged(ScreenWidth,ScreenHeight:integer);
  constructor Create(AnOwner:TComponent);override;
  procedure Clr;
  procedure WriteChar(ch:char);
  procedure BlinkCaret;
  property CaretVisible:boolean read FCaretVisible write SetCaretVisible;
  property Modified:boolean read FModified;
  property Active:boolean read FActive;
  property Busy:boolean read FBusy;
  property OnStateChanged:TConsoleStateChangedEvent read FOnStateChanged write FOnStateChanged;
 end;

 destructor TConsole.Destroy;
 begin
  inherited;
  if Assigned(FScreen) then Dispose(FScreen)
 end;

 function TConsole.Line;
 begin
  SetLength(Result,COLS);
  Move(FScreen^[{$IFDEF SCREENBUF}bufRow+{$ENDIF}r],Result[1],COLS)
 end;

 procedure TConsole.FontChanged;
 var
  DC: HDC;
  Save: THandle;
  Metrics: TTextMetric;

  procedure SetMetrics(const Metrics: TTextMetric);
  begin
    with Metrics do
    begin
      FCharSize.X := tmAveCharWidth;
      FCharSize.Y := tmHeight + tmExternalLeading;
      FOrigin.X:=(ScreenWidth-(COLS*FCharSize.X)) div 2;
      FOrigin.Y:=(ScreenHeight-(ROWS*FCharSize.Y)) div 2
    end;
  end;

 begin
  DC := GetDC(0);
  Save := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);
  SetMetrics(Metrics);
 end;

 procedure TConsole.DrawCaret;
 begin
  OpenScreen;
  {$IFDEF SCREENBUF}
  if BASEBUFROW+row>bufRow+MAXROW then exit;
  {$ENDIF}
  Canvas.Font:=Font;
  if invert then begin
   Canvas.Brush.Color:=Canvas.Font.Color;
   Canvas.Font.Color:=Color
  end else begin
   Canvas.Brush.Color:=Color;
  end;
  Canvas.TextOut(FOrigin.X+col*FCharSize.X,FOrigin.Y+({$IFDEF SCREENBUF}BASEBUFROW-bufRow+{$ENDIF}row)*FCharSize.Y,FScreen^[{$IFDEF SCREENBUF}bufRow-(BASEBUFROW-bufRow)+{$ENDIF}row][col])
 end;

 procedure TConsole.HideCaret;
 begin
  if CaretVisible then DrawCaret(false);
 end;

 procedure TConsole.ShowCaret;
 begin
  if CaretVisible then DrawCaret(FCaretState)
 end;

 procedure TConsole.BlinkCaret;
 begin
  FCaretState:=not FCaretState;
  ShowCaret
 end;

 procedure TConsole.SetActive;
 begin
  if Active<>value then begin
   FActive:=value;
   StateChanged(immediately)
  end
 end;

 procedure TConsole.SetModified;
 begin
  if Modified<>value then begin
   FModified:=value;
   StateChanged(immediately)
  end
 end;

 procedure TConsole.SetBusy;
 begin
  if Busy<>value then begin
   FBusy:=value;
   StateChanged(immediately)
  end
 end;

 procedure TConsole.StateChanged;
 begin
  if Assigned(FOnStateChanged) then FOnStateChanged(Self,repaint)
 end;

 procedure TConsole.SetCaretVisible;
 begin
  if value=CaretVisible then exit;
  if value then begin
   FCaretVisible:=true;
   ShowCaret
  end else begin
   HideCaret;
   FCaretVisible:=false
  end
 end;

 procedure TConsole.OpenScreen;
 begin
  if Assigned(FScreen) then exit;
  New(FScreen);
  FillChar(FScreen^,sizeof(TScreen),' ');
 end;

 constructor TConsole.Create;
 begin
  inherited Create(AnOwner);
  Cursor:=crNone;
// TODO:   DoubleBuffered:=true;
  FBusy:=true;
  {$IFDEF SCREENBUF}
  bufRow:=BASEBUFROW;
  minBufRow:=BASEBUFROW;
  {$ENDIF}
 end;

 procedure TConsole.Clr;
 begin
  OpenScreen;
  FillChar(FScreen^,sizeof(TScreen),' ');
  row:=0;
  col:=0;
  {$IFDEF SCREENBUF}
  bufRow:=BASEBUFROW;
  minBufRow:=BASEBUFROW;
  {$ENDIF}
  Invalidate
 end;

 { TUNE: currently not using scrollwindow looks faster.. }
 { $DEFINE USESCROLLWINDOW}

 // TODO: use flagWrap in painging and return it from the Line() function
 const flagWrap='>'; // TUNE:

 procedure TConsole.WriteChar;
 var i:integer;
 var vis:boolean;

  procedure InvalidateRow(r:integer);
  begin
   if not vis then exit;
   {$IFDEF SCREENBUF}
   inc(r,BASEBUFROW);
   if (r<bufRow) or (r>bufRow+MAXROW) then exit;
   dec(r,bufRow);
   {$ENDIF}
   Canvas.Font := Font;
   Canvas.Brush.Color := Color;
   Canvas.TextOut(FOrigin.X,FOrigin.Y+r*FCharSize.Y { TODO: },Line(r));
  end;

  procedure PC(ch:char);
  begin
   FScreen^[{$IFDEF SCREENBUF}BASEBUFROW+{$ENDIF}row][col]:=ch;
  end;

 label nr;
 var r:TRect;
 label quit;
 begin
  OpenScreen;
  vis:=Tag=MainForm.ActiveMachine;
  if vis then HideCaret;
  case ch of
  #13: col:=0;
  #12: begin row:=0; col:=0 end;
  #31: Clr;
  #10: begin
        nr:
        inc(row);
        if row>MAXROW then begin
         Move(FScreen^[1],FScreen^[0],MAXBUFFERROW*(COLS+ROWFLAGS));
         FillChar(FScreen^[MAXBUFFERROW],COLS+ROWFLAGS,' ');
         dec(row);

         {$IFDEF SCREENBUF}
         if minBufRow<>0 then dec(minBufRow);
         {$ENDIF}

         if vis then begin
         {$IFDEF USESCROLLWINDOW}
         {$IFDEF SCREENBUF}
         if bufRow<BASEBUFROW then begin
          if bufRow>0 then dec(bufRow);
          Invalidate
         end else
         {$ENDIF}
         begin
          r:=ClientRect;
          r.Top:=FOrigin.Y;
          ScrollWindow(Handle,0,-FCharSize.Y { TODO: },@r,nil);
          InvalidateRow(row)
         end
         {$ELSE}
         {$IFDEF SCREENBUF}
         if (bufRow<BASEBUFROW) and (bufRow>0) then dec(bufRow);
         {$ENDIF}
         Invalidate
         {$ENDIF}
         end
        end else
         InvalidateRow(row-1)
       end;
  #7: MessageBeep(0);
  #8: if col>0 then begin dec(col); PC(' '); InvalidateRow(row); goto quit end;
  #9: begin
       PC(' ');
       inc(col);
       if col>MAXCOL then begin
        PC(flagWrap);
        col:=0;
        goto nr
       end;
       while (col mod 8)<>0 do begin
        PC(' ');
        inc(col);
        if col>MAXCOL then begin
         PC(flagWrap);
         col:=0;
         goto nr
        end;
       end;
       InvalidateRow(row)
      end;
  else
   PC(ch);
   inc(col);
   if col>MAXCOL then begin
    PC(flagWrap);
    col:=0;
    goto nr
   end else
    InvalidateRow(row)
  end;
 quit:
  if vis then ShowCaret;
  if not Modified then if not vis then SetModified(true,true)
 end;

 const copr='IQ 30.02'; // TUNE:

 procedure TConsole.PaintScrollers;
 {$IFDEF SCREENBUF}
 var oldColor:TColor;
 var s:string;
 begin
  Canvas.Font := Font;
  Canvas.Brush.Color := Color;

  if minBufRow<bufRow then s:='Up' else s:='  ';
  if bufRow<BASEBUFROW then s:=s+'Dn' else s:=s+'  ';

  oldColor:=Canvas.Font.Color;
  Canvas.Font.Color:=clGray;
  Canvas.TextOut(2 { TUNE: },2 { TUNE: },s);
  Canvas.Font.Color:=oldColor;
 end;
 {$ELSE}
 begin
 end;
 {$ENDIF}

 procedure TConsole.Paint;
 var r:integer;
 var oldColor:TColor;
 var scroll:string;
 begin
// TODO:  PaintScrollers;

  Canvas.Font := Font;
  Canvas.Brush.Color := Color;

  oldColor:=Canvas.Font.Color;
  Canvas.Font.Color:=clGray;

  scroll:='Scroll'; // TUNE:
  if not FScrollLock then FillChar(scroll[1],length(scroll),' ');
  Canvas.TextOut(2 { TUNE: },2 { TUNE: },scroll);

  Canvas.TextOut(Width-Canvas.TextWidth(copr)-2 { TUNE: },2 { TUNE: },copr);
  Canvas.Font.Color:=oldColor;

  if Assigned(FScreen) then
   for r:=0 to MAXROW do
    Canvas.TextOut(FOrigin.X,FOrigin.Y+r*FCharSize.Y { TODO: },Line(r));
  ShowCaret
 end;

const TabWidth=35; // TUNE:

type
 TTabDraw=class(TPanel)
 protected
  Console:TConsole;
  Dirty:boolean;
 public
  constructor Create(AnOwner:TComponent; AParent:TWinControl; AConsole:TConsole);
  procedure ConsoleChanged(repaint:boolean);
  procedure Paint;override;
 end;

 constructor TTabDraw.Create;
 begin
  inherited Create(AnOwner);
  Console:=AConsole;
  BevelOuter:=bvNone;
  Caption:='';
  Width:=TabWidth;
  Align:=alLeft;
  ParentColor:=true;
  ParentFont:=true;
  MainForm.AddTabDraw(TabWidth);
  Parent:=AParent
 end;

 procedure TTabDraw.ConsoleChanged;
 begin
  Dirty:=true;
  if repaint then Invalidate
 end;

const
 clrRed=0;
 clrGreen=1;
 clrGray=2;

const
 colors:array [boolean] of array [clrRed..clrGray] of TColor=(
  (clMaroon, clGreen, clGray),
  (clRed, clLime, clSilver)
 );

 procedure TTabDraw.Paint;
 var Selected:boolean;
 var R:TRect;
 var Index:integer;

 var x,y:integer;

  procedure Put(const s:string; clr:TColor);
  var oldColor:TColor;
  begin
   oldColor:=Canvas.Font.Color;
   Canvas.Font.Color:=clr;
   Canvas.TextOut(x,y,s);
   inc(x,Canvas.TextWidth(s));
   Canvas.Font.Color:=oldColor
  end;

 var s:string;
 var id:string;

 var modified,ready,paused,busy:boolean;

 const ROWS=2;

 const lb='<';
 const rb='>'; // TUNE:

 var i:integer;
 var ox:integer;
 var dy:integer;
 var clr:TColor;
 begin
  Dirty:=false;

  Canvas.Font := Font;
  Canvas.Brush.Color := Color;

  Index:=Console.Tag;
  Selected:=MainForm.ActiveMachine=Index;
  R:=ClientRect;

  { except the "never touched" these are the state vectors to recognize:
       Active Modified Busy
     0 -      -        -     there is something scratched on the console, but its a history now
     1 -      -        .     ERR
     2 -      -        +     ERR
     3 -      +        -     the console has died, but there is something important drawn
     4 -      +        .     ERR
     5 -      +        +     ERR
     6 +      -        -     waiting for user input
     7 +      -        .     paused
     8 +      -        +     busy
     9 +      +        -     waiting for user input, something important is drawn
    10 +      +        .     paused and something important is drawn
    11 +      +        +     busy and something important is drawn

    Visual encoding follows these rules:
    Modified -> use light color ( states 3,9,10,11 )
    paused or dead -> draw thin rectangle ( states 0,3,7,10 )

    waiting for user -> green ( states 6,9 )
    busy or paused -> red ( states 7,8,10,11 )
    dead -> gray ( states 0,3 )
  }

  x:=R.Left+4 { TUNE: };
  dy:=Canvas.TextHeight('Y');
  y:=R.Top+((Height-dy*ROWS) div 2)+dy;
  s:=MainForm.TabNames[index]; //Tabs.Tabs[index];
  id:=IntToStr(index+1);
//  if length(id)<3 then id:=' '+id;

  if Selected then Put(lb,clGray) else Put(' ',clGray);

  i:=Pos(' ',s);
  if i=0 then begin
   { "never touched" }
//   DrawFrame(clBlack { TUNE: });
   Put(id,clGray); //colors[Selected,clrGray])
   if Selected then Put(rb,clGray) else Put(' ',clGray)
  end else begin
   Delete(s,1,i);

   if s[1]=' ' then begin
    { not Active }
    busy:=false;
    paused:=false;
    ready:=false;
    modified:=s[2]<>' '
   end else begin
    { Active }
    case s[3] of
    ' ': { waiting for user }
         begin
          busy:=false;
          paused:=false;
          ready:=true;
          modified:=s[2]<>' '
         end;

    '.': { paused }
         begin
          busy:=false;
          paused:=true;
          ready:=false;
          modified:=s[2]<>' '
         end;

    '+': { busy }
         begin
          busy:=true;
          paused:=false;
          ready:=false;
          modified:=s[2]<>' '
         end
    end
   end;

//   Flag(modified,clAqua);
   ox:=x;
   if ready then
    clr:=clLime
   else
   if paused then
    clr:=clYellow
   else
   if busy then
    clr:=clRed
   else
    clr:=clGray;

   Put(id,clr);

   if Selected then Put(rb,clGray) else Put(' ',clGray);

   dec(y,dy);
   x:=ox;
   if modified then
    Put('_',clr)
   else
    Put(' ',clGray);

  end
 end;

procedure TMainForm.RestartBlinker;
begin
 CaretTimer.Interval:=GetCaretBlinkTime;
 CaretTimer.Enabled:=true;
end;

const MaxThreads=12; // TODO: this is due to

procedure TMainForm.CheckScrollLock;
var lock:boolean;
var view:TWinControl;
begin
 if query then begin
  lock:=(GetAsyncKeyState(VK_SCROLL) and 1)<>0;
  if FScrollLock=lock then exit;
  FScrollLock:=lock;
 end;
 view:=TWinControl(ActiveView);
 if Assigned(view) then view.Invalidate
end;

procedure TMainForm.FormCreate(Sender: TObject);
var i:integer;
begin
 CheckScrollLock(true);
 Threads:=TList.Create;
 TabNames:=TStringList.Create;

 for i:=1 to MaxThreads do NewMachine('',false,-1);

 RestartBlinker;
 if ParamCount=0 then
  NewMachine('',true,-1)
 else
  for i:=1 to ParamCount do
   NewMachine('LOAD "'+ParamStr(i)+'"'#13#10'RUN',true,-1)
end;

procedure TMainForm.BasTerminated;
var i:integer;
var view:TConsole;
begin
{ for i:=Book.PageCount-1 downto 0 do
  if Book.Pages[i].Tag=integer(Sender) then begin
   Book.Pages[i].Free;
   break
  end;
 if Book.PageCount=0 then Close } // TODO:
 view:=TConsole(TBasicThread(Sender).Tag);
 view.SetActive(false,true);
 i:=Threads.IndexOf(Sender);
 Threads[i]:=nil;
 for i:=Threads.Count-1 downto 0 do
  if Threads[i]<>nil then exit;
 Close
end;

procedure TMainForm.InputKey;
var bas:TBasicThread;
var view:TConsole;
begin
 if index<0 then exit;
 bas:=TBasicThread(Threads[index]);
 if Assigned(bas) then begin
  view:=TConsole(bas.Tag);
  view.CaretVisible:=false;
  view.SetBusy(true,false);
  bas.PushKey(Key);
 end;
 Key:=#0
end;

function TMainForm.ActiveThread;
var index:integer;
begin
 index:=ActiveMachine;
 if (index<0) or (index>=Threads.Count) then begin
  Result:=nil;
  exit
 end;
 Result:=Threads[index]
end;

function TMainForm.ActiveView;
var bas:TBasicThread;
var index:integer;
begin
 index:=ActiveMachine;
 if (index<0) or (index>=Threads.Count) then begin
  Result:=nil;
  exit
 end;
 Result:=TWinControl(Book.Pages.Objects[index+1]).Controls[0];
end;

procedure TMainForm.PauseActiveMachine;
var bas:TBasicThread;
var view:TConsole;
begin
 bas:=TBasicThread(Threads[ActiveMachine]);
 if not Assigned(bas) then exit;
 view:=TConsole(bas.Tag);
 if not view.CaretVisible then
  with bas do begin
   view.SetBusy(Suspended,true);
   Suspended:=not Suspended;
  end
end;

procedure TMainForm.OutputKey;
var view:TConsole;
begin
 view:=TConsole(TBasicThread(Sender).Tag);
 view.WriteChar(Key)
end;

procedure TMainForm.BasIdle;
var view:TConsole;
begin
 view:=TConsole(TBasicThread(Sender).Tag);
 view.CaretVisible:=true;
 view.SetBusy(false,true)
end;

procedure TMainForm.ConsoleStateChanged;
var view:TConsole absolute Sender;
var s:string;
begin
 if view.Active then begin
  s:='+';
  if view.Modified then
   s:=s+'+'
  else
   s:=s+' ';
  if view.Busy then
   s:=s+'+'
  else
  if view.CaretVisible then
   s:=s+' '
  else
   s:=s+'.'
 end else begin
  s:=' ';
  if view.Modified then
   s:=s+'+'
  else
   s:=s+' ';
  s:=s+' '
 end;

 TabNames[view.Tag]:=IntToStr(view.Tag+1)+' '+s;
 TTabDraw(TabNames.Objects[view.Tag]).ConsoleChanged(repaint);
end;

procedure TMainForm.ActivateThread;
begin
 ActiveMachine:=index;
 if ActiveThread=nil then
  NewMachine('',true,index)
end;

function TMainForm.NewMachine;
var sname:string;
var bas:TBasicThread;
var view:TConsole;
var page:TWinControl;
var i:integer;
var reuse:boolean;
begin
 reuse:=false;

 if hint>=0 then begin
  if (hint<Threads.Count) and (Threads[hint]=nil) then begin
   reuse:=true;
   Result:=hint+1
  end
 end;

 if active and not reuse then
  for i:=0 to Threads.Count-1 do
   if Threads[i]=nil then begin
    reuse:=true;
    Result:=i+1;
    break
   end;

 if not reuse then begin
  if Threads.Count=MaxThreads then begin
   { TODO: throw exception ? }
   Result:=-1;
   exit
  end;

  Result:=Book.Pages.Count;
  sname:=IntToStr(Result);
  Book.PageIndex:=Book.Pages.Add(sname);
  page:=TWinControl(Book.Pages.Objects[Result]);

  view:=TConsole.Create(page);
  view.Tag:=Result-1;

  TabNames.AddObject(sname,TTabDraw.Create(Self,Tabs,view));

  view.OnStateChanged:=ConsoleStateChanged;
  view.Color:=clBlack;
  view.Font.Color:=clSilver; // TODO: clLime;
  view.Parent:=page;
  view.Align:=alClient;
{  view.Font.Name:='Courier New';
  view.Font.Size:=10;}
  view.Font.Name:='Fixedsys';

  view.FontChanged(Screen.Width,Screen.Height{-Tabs.Height});
 end else begin
  Book.PageIndex:=Result;
  page:=TWinControl(Book.Pages.Objects[Result]);

  view:=TConsole(page.Controls[0]);
  view.Clr
 end;

 if active then begin
  bas:=TBas.Create(basShell.TBasic,cmd);
  bas.FreeOnTerminate:=true;
  bas.OnTerminate:=BasTerminated;
  bas.Tag:=integer(view);
  bas.OnSendKey:=OutputKey;
  bas.OnIdle:=BasIdle;

  view.SetActive(true,true);

  if reuse then
   Threads[Result-1]:=bas
  else
   Threads.Add(bas);

  bas.Resume;
 end else
  Threads.Add(nil);

 dec(Result)
end;

procedure TMainForm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);

 function K(index:integer; vk:integer):boolean;
 begin
  Result:=(Key=vk) and (Shift=[]);
  if Result then begin
   Key:=0;
   ActiveMachine:=index
  end
 end;

 function N(index:integer; vk:integer):boolean;
 begin
  Result:=(Key=vk) and (Shift=[ssCtrl]);
  if Result then begin
   Key:=0;
   ActivateThread(index)
  end
 end;

begin
 if K(0,VK_F1) or
    K(1,VK_F2) or
    K(2,VK_F3) or
    K(3,VK_F4) or
    K(4,VK_F5) or
    K(5,VK_F6) or
    K(6,VK_F7) or
    K(7,VK_F8) or
    K(8,VK_F9) or
    K(9,VK_F10) or
    K(10,VK_F11) or
    K(11,VK_F12) then exit;

 if N(0,VK_F1) or
    N(1,VK_F2) or
    N(2,VK_F3) or
    N(3,VK_F4) or
    N(4,VK_F5) or
    N(5,VK_F6) or
    N(6,VK_F7) or
    N(7,VK_F8) or
    N(8,VK_F9) or
    N(9,VK_F10) or
    N(10,VK_F11) or
    N(11,VK_F12) then exit;

 if (Key=VK_PAUSE) and (Shift=[]) then begin
  Key:=0;
  PauseActiveMachine
 end else
 if (Key=VK_RETURN) and (Shift=[ssCtrl]) then begin
  Key:=0;
  ActivateThread(ActiveMachine)
 end
end;

procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
 InputKey(ActiveMachine,Key)
end;

function TMainForm.GetActiveMachine;
begin
 Result:=Book.PageIndex-1
end;

procedure TMainForm.SetActiveMachine;
var view:TConsole;
var oldMach:integer;
begin
 if (value<0) or (value>=Threads.Count) then exit;
 oldMach:=ActiveMachine;
 if oldMach=value then exit;
 Book.PageIndex:=value+1;
 view:=TConsole(TWinControl(Book.Pages.Objects[Book.PageIndex]).Controls[0]);
 if oldMach>=0 then TWinControl(TabNames.Objects[oldMach]).Invalidate;
 view.SetModified(false,false);
 TWinControl(TabNames.Objects[value]).Invalidate;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
 Threads.Free;
 TabNames.Free
end;

procedure TMainForm.CaretTimerTimer(Sender: TObject);
var view:TConsole;
var i:integer;
begin
 CaretTimer.Enabled:=false;
 view:=TConsole(ActiveView);
 if Assigned(view) then view.BlinkCaret;
 for i:=TabNames.Count-1 downto 0 do
  with TTabDraw(TabNames.Objects[i]) do
   if Dirty then Invalidate;
 RestartBlinker
end;

procedure TMainForm.FormResize(Sender: TObject);
var i:integer;
begin
 for i:=Book.Pages.Count-1 downto 1 do
  TConsole(TWinControl(Book.Pages.Objects[i]).Controls[0]).FontChanged(Screen.Width,Screen.Height{-Tabs.Height});
end;

procedure TMainForm.AddTabDraw;
var w:integer;
begin
 Tabs.Tag:=Tabs.Tag+incr;
 w:=(Tabs.Width-Tabs.Tag) div 2;
 if w<0 then w:=1;
 TabPad.Width:=w
end;

procedure TMainForm.TabsResize(Sender: TObject);
begin
 AddTabDraw(0)
end;

procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 {$IFDEF SCREENBUF}
 if (Key=VK_SCROLL) and (Shift=[]) then begin
  FScrollLock:=not FScrollLock;
  CheckScrollLock(false)
 end else
 if FScrollLock then begin
  if (Key=VK_UP) and (Shift=[]) then begin
   Key:=0;
   ScrollActiveConsole(-1)
  end else
  if (Key=VK_DOWN) and (Shift=[]) then begin
   Key:=0;
   ScrollActiveConsole(1)
  end else
  if (Key=VK_PRIOR) and (Shift=[]) then begin
   Key:=0;
   ScrollActiveConsole(-MAXROW)
  end else
  if (Key=VK_NEXT) and (Shift=[]) then begin
   Key:=0;
   ScrollActiveConsole(MAXROW)
  end else
  if (Key=VK_HOME) and (Shift=[ssCtrl]) then begin
   Key:=0;
   ScrollActiveConsole(-BUFFERROWS)
  end else
  if (Key=VK_END) and (Shift=[ssCtrl]) then begin
   Key:=0;
   ScrollActiveConsole(BUFFERROWS)
  end
 end 
 {$ENDIF}
end;

procedure TMainForm.ScrollActiveConsole;
var view:TConsole;
var base:integer;
begin
 view:=TConsole(ActiveView);
 if not Assigned(view) then exit;
 base:=view.bufRow+dif;
 if base<view.minBufRow then base:=view.minBufRow;
 if base>BASEBUFROW then base:=BASEBUFROW;
 if base=view.bufRow then exit;
 view.bufRow:=base;
 view.Invalidate
end;

procedure TMainForm.FormActivate(Sender: TObject);
begin
 CheckScrollLock(true)
end;

end.
